@@ -1,7 +1,8 @@
 /* run-parts: run a bunch of scripts in a directory
  *
- * Debian Linux run-parts program, version 0.5
+ * Debian Linux run-parts program
  * Copyright (C) 1996 Jeff Noxon <jeff@router.patch.net>,
+ * Copyright (C) 1996-1998 Guy Maor <maor@debian.org>
  *
  * This is free software; see the GNU General Public License version 2
  * or later for copying conditions.  There is NO warranty.
@@ -18,20 +19,29 @@
  * 06/01/96 - AND - v0.5 - Added --verbose option to print the name of each
  *                         script before executing it
  *
+ * 01/20/98 - GM  - v0.6 - Added --report option.
+ *
  */
 
 #include <stdio.h>
 #include <stdarg.h>
 #include <stdlib.h>
 #include <sys/types.h>
+#include <sys/time.h>
 #include <sys/wait.h>
 #include <dirent.h>
 #include <sys/stat.h>
 #include <unistd.h>
+#include <fcntl.h>
 #include <getopt.h>
+#include <string.h>
+#include <errno.h>
+
+#define VERSION "0.7"
 
 int test_mode = 0;
 int verbose_mode = 0;
+int report_mode = 0;
 int exitstatus = 0;
 
 void error (char *format,...)
@@ -47,40 +57,50 @@ void error (char *format,...)
   fprintf (stderr, "\n");
 }
 
-void version (void)
+
+void version ()
 {
-  fprintf (stderr,
-    "Debian Linux run-parts program, version 0.5.\n"
-    "Distributed with Debian's base system.  Copyright (C) 1996 Jeff Noxon.\n"
-    "This is free software; see the GNU General Public License version 2\n"
-    "or later for copying conditions.  There is NO warranty.\n");
+  fprintf (stderr, "Debian GNU/Linux run-parts program, version " VERSION
+"\nCopyright (C) 1994 Ian Jackson, Copyright (C) 1996 Jeff Noxon.
+Copyright (C) 1996,1997,1998 Guy Maor
+This is free software; see the GNU General Public License version 2
+or later for copying conditions.  There is NO warranty.
+");
   exit (0);
 }
 
-void usage (void)
+
+void usage ()
 {
-  fprintf (stderr,
-    "usage: run-parts [--test|--help|--version|--verbose] [--umask=umask] "
-	   "directory\n");
+  fprintf (stderr, "Usage: run-parts [OPTION]... DIRECTORY
+  --test          print script names which would run, but don't run them.
+  --verbose       print script names before running them.
+  --report        print script names if they produce output.
+  --umask=UMASK   sets umask to UMASK (octal), default is 022.
+  --version       output version information and exit.
+  --help          display this help and exit.
+");
+  exit(0);
 }
 
+
 /* The octal conversion in libc is not foolproof; it will take the 8 and 9
  * digits under some circumstances.  We'll just have to live with it.
  */
-void set_umask (void)
+void set_umask ()
 {
   int mask, result;
 
   result = sscanf (optarg, "%o", &mask);
   if ((result != 1) || (mask > 07777) || (mask < 0)) {
     error ("bad umask value");
-    usage ();
     exit (1);
   }
 
   umask (mask);
 }
 
+
 /* True or false? Is this a valid filename (upper/lower alpha, digits,
  * underscores, and hyphens only?)
  */
@@ -99,22 +119,126 @@ int valid_name (char *filename)
   return 1;
 }
 
+
+void set_fl (int fd, int flags)
+{
+  int val;
+  if ((val = fcntl(fd, F_GETFL, 0)) < 0) {
+    error("fcntl F_GETFL: %s", strerror(errno));
+    exit(1);
+  }
+  val |= flags;
+  if (fcntl(fd, F_SETFL, val) < 0) {
+    error("fcntl F_SETFL: %s", strerror(errno));
+    exit(1);
+  }
+}
+
+
 /* Execute a file */
 void run_part (char *progname)
 {
   int result;
   int pid;
+  int pout[2], perr[2];
+
+  if (report_mode && (pipe(pout) || pipe(perr))) {
+    error("pipe: %s", strerror(errno));
+    exit(1);
+  }
   if ((pid=fork()) < 0) {
-    error ("failed to fork: %s", strerror (errno));
-    exit (1);
+    error("failed to fork: %s", strerror(errno));
+    exit(1);
   }
   else if (!pid) {
+    if (report_mode) {
+      if (dup2(pout[1], STDOUT_FILENO) == -1 ||
+	  dup2(perr[1], STDERR_FILENO) == -1) {
+	error("dup2: %s", strerror(errno));
+	exit(1);
+      }
+      close(pout[0]);
+      close(perr[0]);
+      close(pout[1]);
+      close(perr[1]);
+    }
     execl (progname, progname, 0);
     error ("failed to exec %s: %s", progname, strerror (errno));
     exit (1);
   }
-  else
-    wait (&result);
+
+  if (report_mode) {
+    fd_set set;
+    int max, r, printflag;
+    ssize_t c;
+    char buf[4096];
+    
+    close(pout[1]);
+    close(perr[1]);
+    set_fl(pout[0], O_NONBLOCK);
+    set_fl(perr[0], O_NONBLOCK);
+    max = pout[0] > perr[0] ? pout[0]+1 : perr[0]+1;
+    printflag = 0;
+    for (;;) {
+      if (pout[0] < 0 && perr[0] < 0) {
+	r = waitpid(pid, &result, 0);
+	if (r > 0)
+	  break;
+	else {
+	  error("waitpid: %s", strerror(errno));
+	  exit(1);
+	}
+      }
+
+      FD_ZERO(&set);
+      if (pout[0] >= 0) FD_SET(pout[0], &set);
+      if (perr[0] >= 0) FD_SET(perr[0], &set);
+      r = select(max, &set, 0, 0, 0);
+      if (r > 0) {
+	if (pout[0] >= 0 && FD_ISSET(pout[0], &set)) {
+	  c = read(pout[0], buf, sizeof(buf));
+	  if (c > 0) {
+	    if (!printflag) {
+	      printf("%s:\n", progname);
+	      fflush(stdout);
+	      printflag = 1;
+	    }
+	    write(STDOUT_FILENO, buf, c);
+	  }
+	  else if (c == 0) {
+	    close(pout[0]);
+	    pout[0] = -1;
+	  }
+	}
+	if (perr[0] >= 0 && FD_ISSET(perr[0], &set)) {
+	  c = read(perr[0], buf, sizeof(buf));
+	  if (c > 0) {
+	    if (!printflag) {
+	      fprintf(stderr, "%s:\n", progname);
+	      fflush(stderr);
+	      printflag = 1;
+	    }
+	    write(STDERR_FILENO, buf, c);
+	  }
+	  else if (c == 0) {
+	    close(perr[0]);
+	    perr[0] = -1;
+	  }
+	}
+      }
+      else if (r < 0) {
+	error("select: %s", strerror(errno));
+	exit(1);
+      }
+      else {			/* should never happen */
+	error("internal error: select() returned 0 !");
+	exit(1);
+      }
+    }
+  }
+  else {
+    waitpid(pid, &result, 0);
+  }
   
   if (WIFEXITED (result) && WEXITSTATUS(result)) {
     error ("%s exited with return code %d", progname, WEXITSTATUS(result));
@@ -127,6 +251,7 @@ void run_part (char *progname)
   }
 }
 
+
 /* Find the parts to run & call run_part() */
 void run_parts (char *dirname)
 {
@@ -157,7 +282,7 @@ void run_parts (char *dirname)
 	  printf ("run-parts would run %s\n", filename);
 	else {
 	  if (verbose_mode)
-	    fprintf(stderr, "run-parts: about to run %s...\n", filename);
+	    fprintf(stderr, "run-parts: executing %s\n", filename);
 	  run_part (filename);
 	}
       }
@@ -178,49 +303,45 @@ int main (int argc, char *argv[])
 {
   umask (022);
   
-  while (1) {
+  for (;;) {
     int c;
     int option_index = 0;
 
     static struct option long_options[] = {
-      {"test", 0, 0, 0},
-      {"help", 0, 0, 0},
-      {"version", 0, 0, 0},
-      {"umask", 1, 0, 0},
-      {"verbose", 0, 0, 0},
+      {"test", 0, &test_mode, 1},
+      {"verbose", 0, &verbose_mode, 1},
+      {"report", 0, &report_mode, 1},
+      {"umask", 1, 0, 1},
+      {"help", 0, 0, 2},
+      {"version", 0, 0, 3},
       {0, 0, 0, 0}
     };
 
     opterr = 0;
     c = getopt_long (argc, argv, "", long_options, &option_index);
-    
-    if (c == -1)
+    if (c == EOF)
       break;
-    
-    if (c == 0) {
-      switch (option_index) {
-      case 0:
-	test_mode = 1;
-	break;
-      case 2:
-	version ();
-      case 3:
-	set_umask ();
-	break;
-      case 4:
-	verbose_mode = 1;
-	break;
-      case 1:
-      default:
-	usage ();
-	exit (0);
-      }
+    switch (c) {
+    case 0:
+      break;
+    case 1:
+      set_umask ();
+      break;
+    case 2:
+      usage();
+    case 3:
+      version ();
+    default:
+      error("unrecognized option `%s'", argv[optind]);
+      fprintf(stderr, "Try `run-parts --help' for more information.\n");
+      exit (1);
     }
   }
   
   /* We require exactly one argument: the directory name */
   if (optind != (argc - 1)) {
-    usage ();
+    error("missing operand");
+    fprintf(stderr, "Try run-parts --help' for more information.\n");
     exit (1);
   }
 
